package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import org.littletonrobotics.junction.Logger;

import com.revrobotics.RelativeEncoder;
import com.revrobotics.spark.ClosedLoopSlot;
import com.revrobotics.spark.SparkClosedLoopController;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.ControlType;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.MathUtil;
import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.trajectory.TrapezoidProfile;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ElevatorConstants;

public class Elevator extends SubsystemBase {
    protected SparkMax elevatorMotor1;
    protected SparkMax elevatorMotor2;

    protected RelativeEncoder encoder;

    private DigitalInput bottomLimitSwitch;

    private ElevatorFeedforward feedForward;

    private TrapezoidProfile trapProfile;
    private TrapezoidProfile trapProfileAlgae;
    private TrapezoidProfile.State goal;
    private TrapezoidProfile.State setpoint;

    private SparkClosedLoopController controller;
    
    public Elevator() {
        elevatorMotor1 = new SparkMax(
            ElevatorConstants.kElevatorMotor1ID, 
            MotorType.kBrushless
        );
        
        elevatorMotor2 = new SparkMax(
            ElevatorConstants.kElevatorMotor2ID,
            MotorType.kBrushless
        );

        elevatorMotor1.configure(
            ElevatorConstants.motorConfig, 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        elevatorMotor2.configure(
            ElevatorConstants.motorConfig.follow(ElevatorConstants.kElevatorMotor1ID), 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        encoder = elevatorMotor1.getEncoder();

        bottomLimitSwitch = new DigitalInput(
            ElevatorConstants.kBottomLimitSwitchID
        );

        feedForward = new ElevatorFeedforward(
            ElevatorConstants.kFeedForwardS, 
            ElevatorConstants.kFeedForwardG, 
            ElevatorConstants.kFeedForwardV
        );

        trapProfile = new TrapezoidProfile(new TrapezoidProfile.Constraints(ElevatorConstants.kMaxVelocity, ElevatorConstants.kMaxAcceleration));

        trapProfileAlgae = new TrapezoidProfile(new TrapezoidProfile.Constraints(ElevatorConstants.kMaxVelocityAlgae, ElevatorConstants.kMaxAccelerationAlgae));

        controller = elevatorMotor1.getClosedLoopController();
    }

    @Override
    public void periodic() {
        if (!getBottomLimitSwitch()) {
            encoder.setPosition(0);
        }

        Logger.recordOutput("elevator position", getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to the encoder's current position
     * and the elevator brace position
     * 
     * @return Is the motion safe
     */
    public boolean isMotionSafe() {
        return isMotionSafe(getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to some target position and the elevator
     * brace position
     * 
     * @param motionTarget The target position to determine the safety of
     * @return Is the motion safe
     */
    public boolean isMotionSafe(double motionTarget) {
        return motionTarget > ElevatorConstants.kBracePosition;
    }
    
    /**
     * A manual translation command that uses feed forward calculation to maintain position
     * 
     * @param speed The speed at which the elevator translates
     * @return Sets motor voltage to translate the elevator and maintain position
     */
    public Command runManualElevator(DoubleSupplier speed) {
        return run(() -> {
            double desired = speed.getAsDouble();

            if(Math.abs(MathUtil.applyDeadband(desired, .05)) > 0) {
                elevatorMotor1.set(
                    speed.getAsDouble()
                );
            } else {
                elevatorMotor1.setVoltage(feedForward.calculate(0));
            }
            
        });
    }

    /**
     * A command that will use the feed forward to hold up the elevator.
     * Used for feed forward tuning.
     * 
     * @return Sets motor voltage based on feed forward calculation.
     */
    public Command maintainPosition() {

        return startRun(() -> {

        }, 
        () -> {
            controller.setReference(
                encoder.getPosition(), 
                ControlType.kPosition, 
                ClosedLoopSlot.kSlot0,
                feedForward.calculate(0.0)
            );

        });
        
    }

    public Command homeCommand(){
        return run(() -> {
            elevatorMotor1.setVoltage(0.5);
        })
        .until(() -> elevatorMotor1.getOutputCurrent() > 5)
        .andThen(run(() -> encoder.setPosition(0)));
    }

    /**
     * Moves the elevator to a target destination (setpoint). 
     * 
     * @param setpoint Target destination of the subsystem
     * @return Sets motor voltage to achieve the target destination
     */
    public Command goToSetpoint(DoubleSupplier setGoal) {
        
        return startRun(() -> {
            goal = new TrapezoidProfile.State(setGoal.getAsDouble(), 0.0);
        }, () -> {
            setpoint = trapProfile.calculate(0.02, new TrapezoidProfile.State(encoder.getPosition(), encoder.getVelocity()), goal);

            controller.setReference(
                setpoint.position, 
                ControlType.kPosition, 
                ClosedLoopSlot.kSlot0,
                feedForward.calculate(encoder.getVelocity())
            );

        }).until(() -> trapProfile.isFinished(encoder.getPosition()));

    }          

    public Command goToSetpointAlgae(DoubleSupplier setGoal) {
            
        return startRun(() -> {
            goal = new TrapezoidProfile.State(setGoal.getAsDouble(), 0.0);
        }, () -> {
            setpoint = trapProfileAlgae.calculate(0.02, new TrapezoidProfile.State(encoder.getPosition(), encoder.getVelocity()), goal);

            controller.setReference(
                setpoint.position, 
                ControlType.kPosition, 
                ClosedLoopSlot.kSlot0,
                feedForward.calculate(setpoint.velocity)
            );

        }).until(() -> trapProfileAlgae.isFinished(encoder.getPosition()));
    }   

    /**
     * Returns the current encoder position
     * 
     * @return Current encoder position
     */
    public double getEncoderPosition() {
        return encoder.getPosition();
    }


    /**
     * Returns the value of the bottom limit switch on the elevator (false = disabled, true = enabled)
     * 
     * @return The value of bottomLimitSwitch
     */
    public boolean getBottomLimitSwitch() {
        return bottomLimitSwitch.get();
    }

    /**
     * Returns the motor's output current
     * 
     * @return Motor output current
     */
    public double getMotor1() {
        return elevatorMotor1.getAppliedOutput()*elevatorMotor1.getBusVoltage();
    }

    /**
     * Returns the motor's output current
     * 
     * @return Motor output current
     */
    public double getMotor2() {
        return elevatorMotor2.getAppliedOutput()*elevatorMotor2.getBusVoltage();
    }

}